/* Module      : Camera.cpp
 * Author      : Peter Lohrmann
 * Email       : pjlrhino@wpi.edu
 * Course      : IMGD 3000
 *
 * Description : This is the Jaeger Camera class. It allows the camera
 *               to be reppositioned and oriented by the user.
 *
 * Date        : 2006/03/13
 *
 * History:
 * Revision      Date          Changed By
 * --------      ----------    ----------
 * 00.01         2006/03/13    gogo
 * Modified the formatting.
 *
 * (c) Copyright 2006, Worcester Polytechnic Institute.
 */

/* -- INCLUDE FILES ------------------------------------------------------ */
#ifdef MACOSX
  #include <gl.h>
  #include <glu.h>
  #include <glut.h>
#else
  #include <GL/gl.h>
  #include <GL/glu.h>
  #include <GL/glut.h>
#endif

#include <math.h>
#include <Camera.h>

/* -- DATA STRUCTURES ---------------------------------------------------- */

/* -- GLOBAL VARIABLES --------------------------------------------------- */

/* -- LOCAL VARIABLES ---------------------------------------------------- */
static float PI_DIV_180 = ( M_PI / 180.0 );

/* ----------------------------------------------------------------------- */
/* Function    : void Camera::Camera( void )
 *
 * Description : This is the Camera default constructor.
 *
 * Parameters  : void.
 *
 * Returns     : void.
 */

Camera::Camera( )  {
  ;
}

/* ----------------------------------------------------------------------- */
/* Function    : void Camera::~Camera( void )
 *
 * Description : This is the Camera default destructor.
 *
 * Parameters  : void.
 *
 * Returns     : void.
 */

Camera::~Camera( )  {
  ;
}

/* ----------------------------------------------------------------------- */
/* Function    : void Camera::Roll( float angle )
 *
 * Description : This function applies a roll to the camera of 'angle'
 *               degrees.
 *
 * Parameters  : float angle : Degrees to roll the camera by.
 *
 * Returns     : void.
 */

void Camera::Roll( float angle )  {
  float cs = cos( PI_DIV_180 * angle );
  float sn = sin( PI_DIV_180 * angle );

  Vector3 t = u; // remember old u
  u.set( ( cs * t.x ) - ( sn * v.x ),
         ( cs * t.y ) - ( sn * v.y ),
         ( cs * t.z ) - ( sn * v.z ) );
  v.set( ( sn * t.x ) + ( cs * v.x ),
         ( sn * t.y ) + ( cs * v.y ),
         ( sn * t.z ) + ( cs * v.z ) );
  SetModelViewMatrix( );
}

/* ----------------------------------------------------------------------- */
/* Function    : void Camera::Pitch( float angle )
 *
 * Description : This function applies a pitch to the camera of 'angle'
 *               degrees.
 *
 * Parameters  : float angle : Degrees to pitch the camera by.
 *
 * Returns     : void.
 */

void Camera::Pitch( float angle )  {
  float cs = cos( PI_DIV_180 * angle );
  float sn = sin( PI_DIV_180 * angle );

  Vector3 t = v; // remember old v
  v.set( ( cs * t.x ) - ( sn * n.x ),
         ( cs * t.y ) - ( sn * n.y ),
         ( cs * t.z ) - ( sn * n.z ) );
  n.set( ( sn * t.x ) + ( cs * n.x ),
         ( sn * t.y ) + ( cs * n.y ),
         ( sn * t.z ) + ( cs * n.z ) );
  SetModelViewMatrix( );
}

/* ----------------------------------------------------------------------- */
/* Function    : void Camera::Yaw( float angle )
 *
 * Description : This function applies a yaw to the camera of 'angle'
 *               degrees.
 *
 * Parameters  : float angle : Degrees to yaw the camera by.
 *
 * Returns     : void.
 */

void Camera::Yaw( float angle )  {
  float cs = cos( PI_DIV_180 * angle );
  float sn = sin( PI_DIV_180 * angle );

  Vector3 t = u; // remember old u
  u.set( ( cs * t.x ) - ( sn * n.x ),
         ( cs * t.y ) - ( sn * n.y ),
         ( cs * t.z ) - ( sn * n.z ) );
  n.set( ( sn * t.x ) + ( cs * n.x ),
         ( sn * t.y ) + ( cs * n.y ),
         ( sn * t.z ) + ( cs * n.z ) );
  SetModelViewMatrix( );
}

/* ----------------------------------------------------------------------- */
/* Function    : void Camera::Slide( float delU, float delV, float delN )
 *
 * Description : This function applies a translation to the camera,
 *               moving it 'delU' units along X, 'delV' units along Y, and
 *               'delN' units along Z (in camera space).
 *
 * Parameters  : float delU : Number of units to slide along camera X.
 *               float delV : Number of units to slide along camera Y.
 *               float delN : Number of units to slide along camera Z.
 *
 * Returns     : void.
 */

void Camera::Slide( float delU, float delV, float delN )  {
  eye.x += ( delU * u.x ) + ( delV * v.x ) + ( delN * n.x );
  eye.y += ( delU * u.y ) + ( delV * v.y ) + ( delN * n.y );
  eye.z += ( delU * u.z ) + ( delV * v.z ) + ( delN * n.z );
  SetModelViewMatrix( );
}

/* ----------------------------------------------------------------------- */
/* Function    : void Camera::Set( Point3 eyePt, Point3 interestPt,
 *                                 Vector3 upVec )
 *
 * Description : This function places the camera at point 'eyePt', pointing
 *               at location 'interestPt' with the "up" vector set to
 *               'upVec'.
 *
 * Parameters  : Point3 eyePt      : Position to place the camera.
 *               Point3 interestPt : Point of interest.
 *               Vector3 upVec     : "Up" vector for camera.
 *
 * Returns     : void.
 */

void Camera::Set( Point3 eyePt, Point3 interestPt, Vector3 upVec )  {
  eye.set( eyePt ); // store the given eye position
  
  // Make n
  n.set( ( eye.x - interestPt.x ),
         ( eye.y - interestPt.y ),
         ( eye.z - interestPt.z ) ) ;

  // Make u = up X n
  u.set( upVec.cross( n ) );
  
  n.normalize( );
  u.normalize( );
  
  // Make v = n X u
  v.set( n.cross( u ) );
  SetModelViewMatrix( );
}

/* ----------------------------------------------------------------------- */
/* Function    : void Camera::SetModelViewMatrix( void )
 *
 * Description : This function sets OpenGL's modelview matrix based on the
 *               current camera state.
 *
 * Parameters  : void.
 *
 * Returns     : void.
 */

void Camera::SetModelViewMatrix( void )  {
  float m[16];
  Vector3 eVec( eye.x, eye.y, eye.z ); // vector of eye

  m[0] = u.x;  m[4] = u.y;  m[8]  = u.z;  m[12] = -eVec.dot( u );
  m[1] = v.x;  m[5] = v.y;  m[9]  = v.z;  m[13] = -eVec.dot( v );
  m[2] = n.x;  m[6] = n.y;  m[10] = n.z;  m[14] = -eVec.dot( n );
  m[3] = 0;    m[7] = 0;    m[11] = 0;    m[15] = 1.0;

  // Load this matrix into OpenGL's modelview matrix.
  glMatrixMode( GL_MODELVIEW );
  glLoadMatrixf( m );
}

/* ----------------------------------------------------------------------- */
/* Function    : void Camera::SetShape( float vAng, float asp, float nearD,
 *                                      float farD )  {
 *
 * Description : This function sets OpenGL's projection matrix to have
 *               a viewing angle of 'vAng', aspect ration 'asp', near
 *               clipping plane of 'nearD', and far clipping plane of
 *               'farD'.
 *
 * Parameters  : float vAng  : Viewing angle to use.
 *               float asp   : Aspect ration of the view.
 *               float nearD : Distance to the near clipping plane.
 *               float farD  : Distance to the far clipping plane.
 *
 * Returns     : void.
 */

void Camera::SetShape( float vAng, float asp, float nearD, float farD )  {
  viewAngle = PI_DIV_180 * vAng;
  aspect    = asp;
  nearDist  = nearD;
  farDist   = farD;

  wHeight   = nearDist * tan( viewAngle * 0.5 );
  wWidth    = wHeight * aspect;

  glMatrixMode( GL_PROJECTION );
  glLoadIdentity( );
  gluPerspective( vAng, aspect, nearDist, farDist );
}

/* ----------------------------------------------------------------------- */
